/* 
 DTMaker.java
 Author: Tomokazu ARITA
 update: 2001/11/27
 update: 2001/12/09
 Last Update: 2003/01/04
*/

package hiformed.parser;

//jdk 1.3.x
import java.util.Vector;

//hiformed
import hiformed.grammar.*;

public class DTMaker{
	
	private MarkedGraph mgraph = null;
	private DTNode dt_node = null;
	
	private int id_counter = 0;
	
	
	//HandleŎg
	//ȂĂ
	private static final int WAY_IN_START = 1;
	private static final int WAY_IN_END   = 2;
	private static final int WAY_OV_START = 3;
	private static final int WAY_OV_END   = 4;
	private static final int WAY_LF_START = 5;
	private static final int WAY_LF_END   = 6;
	

	
	//Reduction processp Pointer
	private MGNode in_start;
	private MGNode in_end;
	private MGNode ov_start;
	private MGNode ov_end;
	private MGNode lf_start;
	private MGNode lf_end;
	
	private MGNode handle = null;
	
	private PreDTree pre_dtree = null;
	private DTBlock dtree = null;
	
	private ProductionManager pro_manager = null;
	private PrecedenceTable pre_table = null;
	MGNodeMarkSet node_mark_set = null;
	
	
	public DTMaker(
		ProductionManager pro_manager, 
		PrecedenceTable pre_table,
		MGNodeMarkSet node_mark_set
		){
		this.pro_manager = pro_manager;
		this.pre_table = pre_table;
		this.node_mark_set = node_mark_set;
	}

	
	//Main!!!!!!!!!!!!!!!!!!!!!!
	public void makeDT(MarkedGraph graph){
            
		this.mgraph = makeMGCopy(graph);
		
		MarkedGraph temp_graph = mgraph;
		MGNode temp_node = mgraph.getStartNode();
		while(temp_node.getNodeLabel() != node_mark_set.getStartNodeLabel()){
			initParser();
			temp_graph = subMakeDT(mgraph);
			temp_node = temp_graph.getStartNode();
		}
		
		
		constructDT(pre_dtree);///

	}
	
	private void initParser(){
		in_start = null;
		in_end   = null;
		ov_start = null;
		ov_end   = null;
		lf_start = null;
		lf_end   = null;
	
		handle = null;
	}
	
	private MarkedGraph makeMGCopy(MarkedGraph graph){
		MarkedGraph copy_graph = new MarkedGraph();
		copy_graph.setStartNode(copyNodes(graph.getStartNode()));
		
		return copy_graph;
	}
	
	
	//MarkedTreeΉ( Only Nested , Not Tessellation)
	private MGNode copyNodes(MGNode node){
		MGNode new_node = new MGNode();
		
		new_node.setNodeID(id_counter++);//counter
		new_node.setNodeLabel(node.getNodeLabel());
		new_node.setAttX(node.getAttX());
		new_node.setAttY(node.getAttY());
		new_node.setAttWidth(node.getAttWidth());
		new_node.setAttHeight(node.getAttHeight());
		
		
		if(node.getInEndNode() != null){
			MGNode add_node = copyNodes(node.getInEndNode());
			new_node.setInEndNode(add_node);
			add_node.setInStartNode(new_node);
		}
		
		
		if(node.getOvStartNode() != null){
			MGNode add_node = copyNodes(node.getOvStartNode());
			new_node.setOvStartNode(add_node);
			add_node.setOvEndNode(new_node);
		}
		
		
		if(node.getLfStartNode() != null){
			MGNode add_node = copyNodes(node.getLfStartNode());
			new_node.setLfStartNode(add_node);
			add_node.setLfEndNode(new_node);
		}
		
		return new_node;
	}
        
	
	
	private MarkedGraph subMakeDT(MarkedGraph graph){
		seekForHandle(graph.getStartNode());
		replaceRightWithLeft(graph, handle);
		
		return graph;
	}
		
	private boolean seekForHandle(MGNode node){
		MGNode temp_node;
		
		//Flags
		boolean move_node = false;
		//int precedence;
		
		Vector nodes = new Vector();
		
		node.checkNode();//_`FbN
		
		if( (temp_node = node.getInEndNode()) != null){//ӂ̂H
			if(node.getEdgeFlag(MGEdgeLabelSet.IN, MGNode.END) == false){//ӂmFH
				node.checkEdge(MGEdgeLabelSet.IN, MGNode.END);
				temp_node.checkEdge(MGEdgeLabelSet.IN, MGNode.START);
				move_node = comparePrecedence(temp_node,MGEdgeLabelSet.IN,node, nodes);
				if(move_node == true){ return true; }
			}
		}
		
		
		if( (temp_node = node.getInStartNode()) != null){
			if(node.getEdgeFlag(MGEdgeLabelSet.IN, MGNode.START) == false){
				node.checkEdge(MGEdgeLabelSet.IN, MGNode.START);
				temp_node.checkEdge(MGEdgeLabelSet.IN, MGNode.END);
				move_node = comparePrecedence(node,MGEdgeLabelSet.IN,temp_node, nodes);
				if(move_node == true){ return true; }
			}
		}
		
		
		if( (temp_node = node.getOvEndNode()) != null){//ӂ̂H
			if(node.getEdgeFlag(MGEdgeLabelSet.OV, MGNode.END) == false){//ӂmFH
				node.checkEdge(MGEdgeLabelSet.OV, MGNode.END);
				temp_node.checkEdge(MGEdgeLabelSet.OV, MGNode.START);
				move_node = comparePrecedence(temp_node,MGEdgeLabelSet.OV,node, nodes);
				if(move_node == true){ return true; }
			}
		}
		
		
		if( (temp_node = node.getOvStartNode()) != null){
			if(node.getEdgeFlag(MGEdgeLabelSet.OV, MGNode.START) == false){
				node.checkEdge(MGEdgeLabelSet.OV, MGNode.START);
				temp_node.checkEdge(MGEdgeLabelSet.OV, MGNode.END);
				move_node = comparePrecedence(node,MGEdgeLabelSet.OV,temp_node, nodes);
				if(move_node == true){ return true; }
			}
			
		}
		
		
		if( (temp_node = node.getLfEndNode()) != null){//ӂ̂H
			if(node.getEdgeFlag(MGEdgeLabelSet.LF, MGNode.END) == false){//ӂmFH
				node.checkEdge(MGEdgeLabelSet.LF, MGNode.END);
				temp_node.checkEdge(MGEdgeLabelSet.LF, MGNode.START);
				move_node = comparePrecedence(temp_node,MGEdgeLabelSet.LF,node, nodes);
				if(move_node == true){ return true; }
			}
			
		}
		
		
		if( (temp_node = node.getLfStartNode()) != null){
			if(node.getEdgeFlag(MGEdgeLabelSet.LF, MGNode.START) == false){
				node.checkEdge(MGEdgeLabelSet.LF, MGNode.START);
				temp_node.checkEdge(MGEdgeLabelSet.LF, MGNode.END);
				move_node = comparePrecedence(node,MGEdgeLabelSet.LF,temp_node, nodes);
				if(move_node == true){ return true; }
			}
			
		}
		
		
		
		if(nodes.size() == 0){
			handle = node;
			return false;
		}else if(nodes.size() > 0){
			for(int i=0; i<nodes.size(); i++){
				boolean flag = seekForHandle((MGNode)nodes.elementAt(i));
				if( flag == true){ return true; }
			}
		}
		
		//G[ꍇ܂ŗB
		return false;//͈ӖȂ
	}
	
	
	private boolean comparePrecedence(MGNode start_node, int edge_label, MGNode end_node, Vector nodes){
		
		int precedence = pre_table.accessTable(
			start_node.getNodeLabel(),
			edge_label,
			end_node.getNodeLabel()
		);
		
		
		if(precedence == PrecedenceTable.EdgeOfHandle){
			if(start_node.getNodeFlag() == false){
				nodes.addElement(start_node);
			}else if(end_node.getNodeFlag() == false){
				nodes.addElement(end_node);
			}
			return false;
			
		}else if(precedence == PrecedenceTable.HandleIsLeft){
			if(start_node.getNodeFlag() == false){
				seekForHandle(start_node);
				return true;
			}else{
				return false;
			}
		}else if(precedence == PrecedenceTable.HandleIsRight){
			if(end_node.getNodeFlag() == false){
				seekForHandle(end_node);
				return true;
			}else{
				return false;
			}
		}else if(precedence == PrecedenceTable.HandleIsLeftRight){
			return false;
		}
		
		return false;
	}
	
	private void replaceRightWithLeft(MarkedGraph graph, MGNode handle){
		graph.initAllFlags();
		
		moveEdgesAroundHandle(handle);
		
		handle = searchStartNode(handle);
		Production production = searchProduction(handle);////
		setDTree(handle,production);
		embedLeftHandSideToGraph(graph, production);
		
		
	}
	
	private void moveEdgesAroundHandle(MGNode handle){
		MGNode handle_node   = handle;
		MGNode temp_node     = null;
		
		handle.checkNode();//_`FbN
		
		int precedence;
		
		if((temp_node = handle.getInStartNode()) != null){
			/*int*/ precedence = pre_table.accessTable(
				handle_node.getNodeLabel(),
				MGEdgeLabelSet.IN,
				temp_node.getNodeLabel()
			);
			
			if( precedence == PrecedenceTable.EdgeOfHandle ){
				if(temp_node.getNodeFlag() == false){
					moveEdgesAroundHandle(temp_node);
				}
			}else if( precedence == PrecedenceTable.HandleIsLeft ||
					precedence == PrecedenceTable.HandleIsLeftRight){
				in_start = temp_node;
				handle.setInStartNode(null);
			}
		}
		
		
		if((temp_node = handle.getInEndNode()) != null){
			/*int*/ precedence = pre_table.accessTable(
				temp_node.getNodeLabel(),
				MGEdgeLabelSet.IN,
				handle_node.getNodeLabel()
			);
			
			if( precedence == PrecedenceTable.EdgeOfHandle ){
				if(temp_node.getNodeFlag() == false){
					moveEdgesAroundHandle(temp_node);
				}
			}else if( precedence == PrecedenceTable.HandleIsRight ||
					precedence == PrecedenceTable.HandleIsLeftRight){
				in_end = temp_node;
				handle.setInEndNode(null);
			}
		}
		
		
		if((temp_node = handle.getOvStartNode()) != null){
			/*int*/ precedence = pre_table.accessTable(
				handle_node.getNodeLabel(),
				MGEdgeLabelSet.OV,
				temp_node.getNodeLabel()
			);
			
			if( precedence == PrecedenceTable.EdgeOfHandle ){
				if(temp_node.getNodeFlag() == false){
					moveEdgesAroundHandle(temp_node);
				}
			}else if( precedence == PrecedenceTable.HandleIsLeft ||
					precedence == PrecedenceTable.HandleIsLeftRight){
				ov_start = temp_node;
				handle.setOvStartNode(null);
			}
		}
		
		
		if((temp_node = handle.getOvEndNode()) != null){
			/*int*/ precedence = pre_table.accessTable(
				temp_node.getNodeLabel(),
				MGEdgeLabelSet.OV,
				handle_node.getNodeLabel()
			);
			
			if( precedence == PrecedenceTable.EdgeOfHandle ){
				if(temp_node.getNodeFlag() == false){
					moveEdgesAroundHandle(temp_node);
				}
			}else if( precedence == PrecedenceTable.HandleIsRight ||
					precedence == PrecedenceTable.HandleIsLeftRight){
				ov_end = temp_node;
				handle.setOvEndNode(null);
			}
		}
		
		
		if((temp_node = handle.getLfStartNode()) != null){
			/*int*/ precedence = pre_table.accessTable(
				handle_node.getNodeLabel(),
				MGEdgeLabelSet.LF,
				temp_node.getNodeLabel()
			);
			
			if( precedence == PrecedenceTable.EdgeOfHandle ){
				if(temp_node.getNodeFlag() == false){
					moveEdgesAroundHandle(temp_node);
				}
			}else if( precedence == PrecedenceTable.HandleIsLeft ||
					precedence == PrecedenceTable.HandleIsLeftRight){
				lf_start = temp_node;
				handle.setLfStartNode(null);
			}
		}
		
		
		if((temp_node = handle.getLfEndNode()) != null){
			/*int*/ precedence = pre_table.accessTable(
				temp_node.getNodeLabel(),
				MGEdgeLabelSet.LF,
				handle_node.getNodeLabel()
			);
			
			if( precedence == PrecedenceTable.EdgeOfHandle ){
				if(temp_node.getNodeFlag() == false){
					moveEdgesAroundHandle(temp_node);
				}
			}else if( precedence == PrecedenceTable.HandleIsRight ||
					precedence == PrecedenceTable.HandleIsLeftRight){
				lf_end = temp_node;
				handle.setLfEndNode(null);
			}
		}
		
		
	}
	
	private Production searchProduction(MGNode handle){
		Production production = pro_manager.getProduction();
		
		MarkedGraph handle_graph = new MarkedGraph();
		handle = searchStartNode(handle);///
		handle_graph.setStartNode(handle);///
		
		while(production != null){
			MarkedGraph right_hand_side = production.getRightHandSide();
			if(compareGraph(handle_graph, right_hand_side)){
				break;
			}else{
				production = production.getNext();
			}
		}
		
		return production;		
		
	}
	
	private boolean compareGraph(MarkedGraph handle_graph, MarkedGraph right_hand_side){
		return subCompareGraph(handle_graph.getStartNode(), right_hand_side.getStartNode());
	}
	
	private boolean subCompareGraph(MGNode node1, MGNode node2){
		MGNode node1_next = null;
		MGNode node2_next = null;
		
		boolean flag_1 = false;
		boolean flag_2 = false;
		boolean flag_3 = false;
		
		
		if(node1.getNodeLabel() != node2.getNodeLabel()){
			return false;
		}else{
			
			if((node1_next = node1.getInEndNode()) != null){
				if((node2_next = node2.getInEndNode()) != null){
					flag_1 = subCompareGraph(node1_next,node2_next);
				}else{
					return false;
				}
			}else{
				if((node2_next = node2.getInEndNode()) != null){
					return false;
				}else{
					flag_1 = true;
				}
			}
			
			
			if((node1_next = node1.getOvStartNode()) != null){
				if((node2_next = node2.getOvStartNode()) != null){
					flag_2 = subCompareGraph(node1_next,node2_next);
				}else{
					return false;
				}
			}else{
				if((node2_next = node2.getOvStartNode()) != null){
					return false;
				}else{
					flag_2 = true;
				}
			}
			
			
			if((node1_next = node1.getLfStartNode()) != null){
				if((node2_next = node2.getLfStartNode()) != null){
					flag_3 = subCompareGraph(node1_next,node2_next);
				}else{
					return false;
				}
			}else{
				if((node2_next = node2.getLfStartNode()) != null){
					return false;
				}else{
					flag_3 = true;
				}
			}
			
		}
		
		if(flag_1 == true && flag_2 == true && flag_3 == true){
			return true;
		}
		
		return false;
	}
	
	private void embedLeftHandSideToGraph(MarkedGraph graph, Production production){
	
		//Left_Han_SideƓx̃m[hnh̏ꏊɖߍ
		MGNode left_hand_side = production.getLeftHandSide();
		MGNode node = new MGNode();
		node.setNodeID(id_counter++);//counter
		node.setNodeLabel(left_hand_side.getNodeLabel());
		
		ConnectionRelation cr = production.getCR();
		while(cr != null){
		
			if(cr.edge_label == MGEdgeLabelSet.IN){
				if(cr.in_out == cr.IN){
					if(in_end != null){
						if( (cr.node_label  == MGNodeMarkSet.ALL_LABEL) 
							|| (cr.node_label == in_end.getNodeLabel())){
						
							connectEdge(node,in_end,cr);
						}
					}
				}else if(cr.in_out == cr.OUT){
					if(in_start != null){
						if( (cr.node_label  == MGNodeMarkSet.ALL_LABEL) 
							|| (cr.node_label == in_start.getNodeLabel())){
							
							connectEdge(node,in_start,cr);
						}
					}
				}
			}else if(cr.edge_label == MGEdgeLabelSet.LF){
				if(cr.in_out == cr.IN){
					if(lf_end != null){
						if( (cr.node_label  == MGNodeMarkSet.ALL_LABEL) 
							|| (cr.node_label == lf_end.getNodeLabel())){
							
							connectEdge(node,lf_end,cr);
						}
					}
				}else if(cr.in_out == cr.OUT){
					if(lf_start != null){
						if( (cr.node_label  == MGNodeMarkSet.ALL_LABEL) 
							|| (cr.node_label == lf_start.getNodeLabel())){
						
							connectEdge(node,lf_start,cr);
						}
					}
				}
			}else if(cr.edge_label == MGEdgeLabelSet.OV){
				if(cr.in_out == cr.IN){
					if(ov_end != null){
						if( (cr.node_label  == MGNodeMarkSet.ALL_LABEL) 
							|| (cr.node_label == ov_end.getNodeLabel())){
						
							connectEdge(node,ov_end,cr);
						}
					}
				}else if(cr.in_out == cr.OUT){
					if(ov_start != null){
						if( (cr.node_label  == MGNodeMarkSet.ALL_LABEL) 
							|| (cr.node_label == ov_start.getNodeLabel())){
							
							connectEdge(node,ov_start,cr);
						}
					}
				}
			}
			
			cr = cr.getNext();
		}
		
		//--------------------kani syori---------------------
		if( (node.getInStartNode() == null) &&
			(node.getOvEndNode() == null) &&
			(node.getLfEndNode() == null) ){
			
			graph.setStartNode(node);
		}

	}
	
	private void connectEdge(MGNode node, MGNode add_node, ConnectionRelation cr){
		
		if(cr.change_edge_label == MGEdgeLabelSet.IN){
			if(cr.in_out == cr.IN){
				node.setInEndNode(add_node);
				add_node.setInStartNode(node);
			}else if(cr.in_out == cr.OUT){
				node.setInStartNode(add_node);
				add_node.setInEndNode(node);
			}
		}else if(cr.change_edge_label == MGEdgeLabelSet.OV){
			if(cr.in_out == cr.IN){
				node.setOvEndNode(add_node);
				add_node.setOvStartNode(node);
			}else if(cr.in_out == cr.OUT){
				node.setOvStartNode(add_node);
				add_node.setOvEndNode(node);
			}
		}else if(cr.change_edge_label == MGEdgeLabelSet.LF){
			if(cr.in_out == cr.IN){
				node.setLfEndNode(add_node);
				add_node.setLfStartNode(node);
			}else if(cr.in_out == cr.OUT){
				node.setLfStartNode(add_node);
				add_node.setLfEndNode(node);
			}
		}
		
	}
	
	private MGNode searchStartNode(MGNode node){
		MGNode start_node = node;
		MGNode next;
		
		while(true){
			if((next = start_node.getLfEndNode()) != null){
				start_node = next;
			}else if((next = start_node.getOvEndNode()) != null){
				start_node = next;
			}else if((next = start_node.getInStartNode()) != null){
				start_node = next;
			}else{
				break;
			}
		}
		
		return start_node;
	}
	
	private void setDTree(MGNode handle,Production production){
		DTBlock block = new DTBlock();
		
		setDTNode(handle, block);//
		
		DTNode new_node = new DTNode();
		MGNode left_hand_side = production.getLeftHandSide();
		String new_node_label = node_mark_set.getNodeMark(left_hand_side.getNodeLabel());
		new_node.setNodeID(id_counter);//id_countergĂ̂Œ/////
		new_node.setNodeLabel(new_node_label);
		new_node.setProduction(production);
		new_node.setNextBlock(block);
		block.setParentNode(new_node);
		
		PreDTree new_predtree = new PreDTree();
		new_predtree.node = new_node;
		new_predtree.next = pre_dtree;//
		pre_dtree = new_predtree;
	}	

	
	
	private void setDTNode(MGNode add_node, DTBlock block){
		
		PreDTree corrent_predtree = null;
		PreDTree before_predtree  = null;
		DTNode temp_node;
		
		
		if(pre_dtree != null){
			corrent_predtree = pre_dtree;
			while(corrent_predtree != null){
				temp_node = corrent_predtree.node;
				if(temp_node.getNodeID() == add_node.getNodeID()){
					block.addDTNode(temp_node);
					if(before_predtree != null){
						before_predtree.next = corrent_predtree.next;
					}else{
						pre_dtree = corrent_predtree.next;
					}
					break;
				}else{
					before_predtree  = corrent_predtree;
					corrent_predtree = corrent_predtree.next;
				}
			}
		}
		
		
		if(corrent_predtree == null){
			block.addNewDTNode(
				add_node.getNodeID(),
				node_mark_set.getNodeMark(add_node.getNodeLabel()),
				add_node.getAttX(),
				add_node.getAttY(),
				add_node.getAttWidth(),
				add_node.getAttHeight()
			);
		}
		
		MGNode next;
		if( (next = add_node.getInEndNode()) != null ){
			setDTNode(next, block);
		}
		if( (next = add_node.getLfStartNode()) != null ){
			setDTNode(next, block);
		}
		if( (next = add_node.getOvStartNode()) != null ){
			setDTNode(next, block);
		}
		
	}
	
	
	
	
	private void constructDT(PreDTree pre_dtree){
		DTNode node = pre_dtree.node;
		dtree = new DTBlock(node);
	}
	
	public DTBlock getDTree(){
		return dtree;
	}
	
	public void evaluteAttribute(DTBlock block){
		subEvaluteAttribute(block.getStartNode());
	}
	
	public void subEvaluteAttribute(DTNode node){
		DTNode  left  = node;
		DTBlock right = node.getNextBlock();
		Production production = null;
		SemanticRule semantic_rule = null;
		if(right == null){
			return;
		}else{
			production = node.getProduction();
			semantic_rule = production.getSemanticRule();
		}
		
		int counter = 1;
		
		
		
		DTNode temp_node = right.getNode(counter);;
		while(temp_node != null){
			semantic_rule.Inh(left, right, counter);
			subEvaluteAttribute(temp_node);
			semantic_rule.Syn(left, right);
			
			counter++;
			temp_node = right.getNode(counter);
		}
		
	}
        
        public void clear(){
            this.mgraph = null;
            this.dtree = null;
            
            id_counter = 0;
            handle = null;
            pre_dtree = null;
            
            pro_manager = null;
            pre_table = null;
            node_mark_set = null;
        }
}
